#include "mt7663_shim.h"
#include "mdp.h"
#include <net/cfg80211.h>
#include <wireless/core.h>

static inline struct mt7663_info *mt7663_get_shim_priv(struct simple80211_info *info)
{
    return ((struct mt7663_info *)info->shim.priv);
}

static inline struct mt7663_lmac_ops *mt7663_get_lmac_ops(struct simple80211_info *info)
{
    return mt7663_get_shim_priv(info)->lmac_ops;
}

static int mt7663_set_channel(struct simple80211_info *info, unsigned int channel) {
    /* FIXME: we may need to involve MTK to implement this. */
    return -EOPNOTSUPP;
}

static int mt7663_get_beacon_capabilities(struct simple80211_info *info, unsigned int *beacon_capabilities)
{
    return mt7663_get_lmac_ops(info)->get_beacon_capabilities(info->base_dev->ieee80211_ptr->wiphy, beacon_capabilities);
}

static int mt7663_get_beacon_interval(struct simple80211_info *info, unsigned int *beacon_interval)
{
    return mt7663_get_lmac_ops(info)->get_beacon_interval(info->base_dev->ieee80211_ptr->wiphy, beacon_interval);
}

static int mt7663_get_channel(struct simple80211_info *info, unsigned int *channel)
{
    return mt7663_get_lmac_ops(info)->get_channel(info->base_dev->ieee80211_ptr->wiphy, channel);
}

static int mt7663_get_link_stats(struct simple80211_info *info,
                                 struct simple80211_link_stats_info *pInfoLocal,
                                 unsigned long thresholdTxPktsMin,
                                 unsigned long thresholdLatencyMs)
{
    return mt7663_get_lmac_ops(info)->get_link_stats(info->base_dev->ieee80211_ptr->wiphy,
                                                     pInfoLocal,
                                                     thresholdTxPktsMin,
                                                     thresholdLatencyMs);
}

static int mt7663_link_quality_polling(struct simple80211_info *info, unsigned int enable)
{
    return mt7663_get_lmac_ops(info)->link_quality_polling(info->base_dev->ieee80211_ptr->wiphy, enable);
}

static void mt7663_config_coex_optimization(struct simple80211_info *info)
{
    return mt7663_get_lmac_ops(info)->config_coex_optimization(info->base_dev->ieee80211_ptr->wiphy);
}

static int mt7663_scan(struct simple80211_info *info, struct simple80211_scan_request *shim_req)
{
    /* Not implemented */
    return -EOPNOTSUPP;
}

static int mt7663_scan_start(struct simple80211_info *info, unsigned long *args)
{
    struct cfg80211_scan_request *pRequest;
    int iRC;

    if (!info->valid) {
        return -EINVAL;
    }

    if (args[2] > IEEE80211_MAX_SSID_LEN) {
        return -EINVAL;
    }

    spin_lock_bh(&info->info_lock);
    if (info->scan_info.scan_flags & SCAN_FLAG_RUNNING) {
        printk("%s: Scan is already running: %d\n", __func__, info->scan_info.scan_flags);
        spin_unlock_bh(&info->info_lock);
        return -EINVAL;
    }
    spin_unlock_bh(&info->info_lock);

    /* This memory block would get freed by cfg80211 when the scan is done or aborted. */
    pRequest = (struct cfg80211_scan_request *) kzalloc(sizeof(*pRequest) + sizeof(*pRequest->ssids), GFP_KERNEL);

    if (!pRequest) {
        printk("%s: Out OF memory!\n", __func__);
        return -ENOMEM;
    }
    pRequest->ssids = (struct cfg80211_ssid *)((void *)pRequest+sizeof(*pRequest));
    if (copy_from_user(pRequest->ssids->ssid, (void *)args[1], args[2])) {
        kfree(pRequest);
        return -EFAULT;
    }

    pRequest->ssids->ssid_len = args[2];
    pRequest->n_ssids = 1;
    pRequest->wdev = info->base_dev->ieee80211_ptr;
    pRequest->wiphy = info->base_dev->ieee80211_ptr->wiphy;
    wiphy_to_rdev(pRequest->wiphy)->scan_req = pRequest;
    pRequest->scan_start = jiffies;
    iRC = mt7663_get_lmac_ops(info)->scan(info->base_dev->ieee80211_ptr->wiphy, pRequest);
    if (!iRC) {
        spin_lock_bh(&info->info_lock);
        info->scan_info.scan_flags |= SCAN_FLAG_RUNNING | SCAN_FLAG_UMAC_SCAN;
        spin_unlock_bh(&info->info_lock);
    } else {
        kfree(pRequest);
    }
    return iRC;
}

static inline int mt7663_scan_abort(struct simple80211_info *info)
{
    mt7663_get_lmac_ops(info)->scan_abort(info->base_dev->ieee80211_ptr->wiphy, info->base_dev->ieee80211_ptr);
    spin_lock_bh(&info->info_lock);
    info->scan_info.scan_flags = 0;
    spin_unlock_bh(&info->info_lock);
    return 0;
}

static inline int mt7663_scan_save_results(struct simple80211_info *info)
{
    mt7663_get_lmac_ops(info)->scan_save_results(info->base_dev->ieee80211_ptr->wiphy);
    return 0;
}

static inline int mt7663_scan_clear(struct simple80211_info *info)
{
    mt7663_get_lmac_ops(info)->scan_clear(info->base_dev->ieee80211_ptr->wiphy);
    return 0;
}

static int mt7663_send_stat_event(struct simple80211_info *info,
				  enum stat_event_ID_t stat_event_ID,
				  unsigned int count)
{
    return mt7663_get_lmac_ops(info)->send_stat_event(info->base_dev->ieee80211_ptr->wiphy,
                                                      stat_event_ID,
                                                      count);
}

#if defined(SONOS_ARCH_ATTR_SUPPORTS_SUSPEND) && defined(SONOS_ARCH_ATTR_SUPPORTS_WOSSID)
static void mt7663_start_wossid(struct simple80211_info *info)
{
    if (mt7663_get_lmac_ops(info)->sched_scan_start) {
        mt7663_get_lmac_ops(info)->sched_scan_start(info->base_dev->ieee80211_ptr->wiphy,
                                                    &info->wossid_list[0],
                                                    info->wossid_list_len);
    }
}

static void mt7663_stop_wossid(struct simple80211_info *info)
{
    if (mt7663_get_lmac_ops(info)->sched_scan_stop) {
        mt7663_get_lmac_ops(info)->sched_scan_stop(info->base_dev->ieee80211_ptr->wiphy);
    }
}
#endif

#if defined(SONOS_ARCH_ATTR_SUPPORTS_SUSPEND)
static void mt7663_pwr_resume(struct simple80211_info *info)
{
    if (mt7663_get_lmac_ops(info)->pwr_resume) {
        mt7663_get_lmac_ops(info)->pwr_resume(info->base_dev->ieee80211_ptr->wiphy);
    }
}

static void mt7663_pwr_suspend(struct simple80211_info *info)
{
    if (mt7663_get_lmac_ops(info)->pwr_suspend) {
        mt7663_get_lmac_ops(info)->pwr_suspend(info->base_dev->ieee80211_ptr->wiphy);
    }
}
#endif

static inline void mt7663_credcheck_reset(struct simple80211_info *info)
{
    mt7663_get_lmac_ops(info)->scan_results_credcheck_reset(info->base_dev->ieee80211_ptr->wiphy);
}

/* Functions provided by the shim to allow the UMAC to call into it. */
static const struct simple80211_shim_ops shim_ops = {
    .get_beacon_capabilities = mt7663_get_beacon_capabilities,
    .get_beacon_interval = mt7663_get_beacon_interval,
    .get_channel = mt7663_get_channel,
    .get_link_stats = mt7663_get_link_stats,
    .link_quality_polling = mt7663_link_quality_polling,
    .scan = mt7663_scan,
    .scan_abort = mt7663_scan_abort,
    .scan_clear = mt7663_scan_clear,
    .scan_save_results = mt7663_scan_save_results,
    .scan_start = mt7663_scan_start,
    .send_stat_event = mt7663_send_stat_event,
    .set_channel = mt7663_set_channel,
#if defined(SONOS_ARCH_ATTR_SUPPORTS_SUSPEND) && defined(SONOS_ARCH_ATTR_SUPPORTS_WOSSID)
    .start_wossid = mt7663_start_wossid,
    .stop_wossid = mt7663_stop_wossid,
#endif
#if defined(SONOS_ARCH_ATTR_SUPPORTS_SUSPEND)
    .pwr_resume = mt7663_pwr_resume,
    .pwr_suspend = mt7663_pwr_suspend,
#endif
    .credcheck_reset = mt7663_credcheck_reset,
};

struct {
    uint8_t mode;
    uint8_t sonos_region;
    char *country_code;
} mt7663_region_map[] = {
    /* Based on different modes, we may want to select different region string
     * in order to have different channel sets. The channel sets corresponding
     * to the country code is defined in reg_rule.c */

    /* NOTE: keep US region as the first and default */
    {SIMPLE80211_OP_MODE_NORMAL, MDP_REGION_USCA, "US"},
    {SIMPLE80211_OP_MODE_NORMAL, MDP_REGION_EU, "TR"},
    {SIMPLE80211_OP_MODE_NORMAL, MDP_REGION_CHINA, "CN"},
    {SIMPLE80211_OP_MODE_NORMAL, MDP_REGION_JAPAN, "JP"},
    {SIMPLE80211_OP_MODE_NORMAL, MDP_REGION_ISRAEL, "IL"},
    {SIMPLE80211_OP_MODE_NORMAL, MDP_REGION_RUSSIA, "RU"},
    {SIMPLE80211_OP_MODE_NORMAL, MDP_REGION_SOUTH_KOREA, "JP"},
};

int mt7663_get_region_string(struct simple80211_info *info, char *country) {

    int idx;
    for (idx = 0; idx < ARRAY_SIZE(mt7663_region_map); idx++) {
        if (mt7663_region_map[idx].mode == info->operation_mode &&
            mt7663_region_map[idx].sonos_region == sys_mdp.mdp_region ) {
            break;
        }
    }

    if (idx >= ARRAY_SIZE(mt7663_region_map)) {
       idx = 0;
       printk("ASSERT: unknown region code %u in mdp!\n", sys_mdp.mdp_region);
    }

    memcpy(country, mt7663_region_map[idx].country_code, 2);

    printk("MTK7663 set country code %s\n", mt7663_region_map[idx].country_code);

    return 0;
}

int mt7663_shim_handle_cmd(struct simple80211_info *info, struct mt7663_shim_cmd *cmd)
{
    int ret = -1;

    if (cmd == NULL || info == NULL) {
        printk("%s: error input info: %p cmd: %p\n", __func__, info, cmd);
        dump_stack();
        return ret;
    }

    switch (cmd->s_cmd.cmd_type) {
        case SIMPLE80211_INIT_INFO:
        case SIMPLE80211_IOCTL:
        case SIMPLE80211_ADD_KEY:
        case SIMPLE80211_SCAN_EVENT:
        case SIMPLE80211_TXRX_FILTER:
        case SIMPLE80211_STA_GET_STATE:
#ifdef SONOS_ARCH_ATTR_SUPPORTS_SUSPEND
        case SIMPLE80211_SUSPEND:
        case SIMPLE80211_RESUME:
#endif
        {
            ret = simple80211_handle_cmd(info, &cmd->s_cmd);
            break;
        }
        case SIMPLE80211_BR_UPLINK:
        {
            ret = simple80211_handle_cmd(info, &cmd->s_cmd);
            mt7663_config_coex_optimization(info);
            break;
        }
        case MT7663_CMD_GET_REGION:
        {
            ret = mt7663_get_region_string(info, cmd->mt7663_data.region.country);
            break;
        }
        default:
            printk("%s: Unknown command type %u\n", __func__, cmd->s_cmd.cmd_type);
    }
    return ret;
}
EXPORT_SYMBOL(mt7663_shim_handle_cmd);

void shim_init(struct simple80211_info *info, void *lmac_ops, int init)
{
    if (init) {
        /* Initialize the shim. */
        struct mt7663_info *shim_info;
        info->shim.shim_ops = &shim_ops;
        /* For mt7663, this is called in sdio probe, where sleep is not allowed */
        shim_info = (struct mt7663_info *)kzalloc(sizeof(struct mt7663_info),
                                                         GFP_ATOMIC);
        if (shim_info == NULL) {
            printk("%s: failed to allocate shim for mt7663\n", __func__);
            return;
        }
        shim_info->lmac_ops = (struct mt7663_lmac_ops *)lmac_ops;
        info->shim.priv = shim_info;
    } else if (info->shim.priv) {
        /* Free up the shim. */
        kfree(info->shim.priv);
    } else {
        printk("%s: ASSERT - simple80211 trying to free up NULL shim\n", __func__);
    }
}

