#ifndef _SHIM_MT7663_H_
#define _SHIM_MT7663_H_

#include "simple80211.h"

#define DEFAULT_FIXED_NOISE_FLOOR -98 /* in dBm */

struct wiphy;
struct cfg80211_scan_request;
struct wireless_dev;

/* Interfaces provided by the LMAC to allow the shim to call into it. */
struct mt7663_lmac_ops {
    void (*config_coex_optimization)(struct wiphy *wiphy);
    int (*get_beacon_capabilities)(struct wiphy *wiphy, unsigned int *beacon_capabilities);
    int (*get_beacon_interval)(struct wiphy *wiphy, unsigned int *beacon_interval);
    int (*get_channel)(struct wiphy *wiphy, unsigned int *channel);
    int (*get_link_stats)(struct wiphy *wiphy, struct simple80211_link_stats_info *pInfoLocal,
                          unsigned long thresholdTxPktsMin, unsigned long thresholdLatencyMs);
    int (*link_quality_polling)(struct wiphy *wiphy, unsigned int enable);
#ifdef SONOS_ARCH_ATTR_SUPPORTS_SUSPEND
    void (*pwr_resume)(struct wiphy *wiphy);
    void (*pwr_suspend)(struct wiphy *wiphy);
#endif
    int (*scan)(struct wiphy *wiphy, struct cfg80211_scan_request *request);
    void (*scan_abort)(struct wiphy *wiphy, struct wireless_dev *wdev);
    void (*scan_clear)(struct wiphy *wiphy);
    void (*scan_results_credcheck_reset)(struct wiphy *wiphy);
    void (*scan_save_results)(struct wiphy *wiphy);
    int (*sched_scan_start)(struct wiphy *wiphy, struct simple80211_ssid *ssids, u_int8_t ssid_num);
    int (*sched_scan_stop)(struct wiphy *wiphy);
    int (*send_stat_event)(struct wiphy *wiphy, enum stat_event_ID_t stat_event_ID, unsigned int count);
    uint8_t (*translate_resume_reason)(uint8_t);
};

struct mt7663_info {
    struct mt7663_lmac_ops *lmac_ops;
};

/* Shim layer specific commands, which start from SIMPLE80211_CMD_END */
enum mt7663_cmd_types {
    MT7663_CMD_GET_REGION = SIMPLE80211_CMD_END,
};

struct mt7663_cmd_get_region {
    char country[2];
};

/* Commands can be issued from LMAC to shim. For common commands, shim can
 * pass them to simple80211 layer to handle, or do some specific handling by
 * itself. Shim specific commands are supposed to be handled by shim layer only.
 */
struct mt7663_shim_cmd {
    struct simple80211_cmd s_cmd;
    union {
        struct mt7663_cmd_get_region region;
    } mt7663_data;
};

void shim_init(struct simple80211_info *info, void *lmac_ops, int init);
extern int mt7663_shim_handle_cmd(struct simple80211_info *info,
                                    struct mt7663_shim_cmd *cmd);
#endif

