/* Copyright (c) 2019-2021, Sonos, Inc.
 *
 * SPDX-License-Identifier: GPL-2.0
 *
 */
#ifndef _SIMPLE80211_IOCTL_H_
#define _SIMPLE80211_IOCTL_H_

#define SIMPLE80211_CTL_START 0

enum simple80211_ctl {
    SIMPLE80211_CTL_SET_HHID = SIMPLE80211_CTL_START,
    SIMPLE80211_CTL_SET_CHANNEL,
    SIMPLE80211_CTL_GET_CHANNEL,
    SIMPLE80211_CTL_STA_SET_ENABLE,
    SIMPLE80211_CTL_ADD_TXRX_FILTER,
    SIMPLE80211_CTL_STA_GET_STATE,
    SIMPLE80211_CTL_GET_OPEN_MODE,
    SIMPLE80211_CTL_CACHE_OPEN_MODE,
    SIMPLE80211_CTL_SET_IP_ADDR,
    SIMPLE80211_CTL_SCAN_START,
    SIMPLE80211_CTL_SCAN_ABORT,
    SIMPLE80211_CTL_SCAN_CLEAR,
    SIMPLE80211_CTL_SCAN_GET_RESULTS,
    SIMPLE80211_CTL_HAS_WIRELESS_CNX,
    SIMPLE80211_CTL_GET_RESUME_REASON,
    SIMPLE80211_CTL_WOSSID_ADD,
    SIMPLE80211_CTL_WOSSID_CLR,
    SIMPLE80211_CTL_GET_LINK_STATS,
};

#define FILTER_ETHERTYPE  (1 << 0)
#define FILTER_IPV4       (1 << 1)
#define FILTER_IPV6       (1 << 2)
#define FILTER_IPPROTO    (1 << 3)
#define FILTER_SPORT      (1 << 4)
#define FILTER_DPORT      (1 << 5)
#define FILTER_MATCH      (1 << 6)
#define FILTER_MANGLE     (1 << 7)
#define FILTER_ALLOW      (1 << 8)
#define FILTER_DROP       (1 << 9)
#define FILTER_MIRROR     (1 << 10)
#define FILTER_TX         (1 << 11)
#define FILTER_RX         (1 << 12)
#define FILTER_TXRX       (FILTER_TX | FILTER_RX)

/* struct txrx_data_pattern used for matching and mangling operates on
 * following parts of given packet:
 * - For TCP/UDP packet, operate on the part after TCP/UDP header.
 * - For IP packet that’s not TCP/UDP, operate on the part after IP header.
 * - For packet that’s not a IP packet, operate on the part after ethernet header.
 */
struct txrx_data_pattern {
    u_int16_t offset;
    u_int8_t mask;
    u_int8_t value;
};

struct simple80211_txrx_filter {
    u_int32_t flags;
    u_int16_t ethertype;
    u_int8_t ipproto;
    u_int16_t dport;
    u_int16_t sport;
    struct txrx_data_pattern match;
    struct txrx_data_pattern mangle;
};

#define MAX_NUM_APS_IN_SCAN_RESULTS 255
#ifndef IEEE80211_ADDR_LEN
#define IEEE80211_ADDR_LEN 6
#endif
#ifndef IEEE80211_MAX_SSID_LEN
#define IEEE80211_MAX_SSID_LEN 32
#endif
#ifndef IEEE80211_MAX_HHID_LEN
#define IEEE80211_MAX_HHID_LEN 32
#endif

struct simple80211_scan_results_ap {
    u_int8_t  bssid[IEEE80211_ADDR_LEN];
    u_int8_t  ssid[IEEE80211_MAX_SSID_LEN];
    u_int8_t  ssid_len;
    u_int8_t  rssi;
    u_int8_t  channel;
    u_int64_t flags;
    u_int32_t sys_ts;
};

struct simple80211_scan_results {
    struct simple80211_scan_results_ap aps[MAX_NUM_APS_IN_SCAN_RESULTS];
    u_int16_t num_aps;
    u_int8_t active;
    unsigned long timestamp;
    time_t rtc_ts;
    u_int32_t sys_ts;
};

/* Copied and extended from atheros/driver/ath_ioctl.h */
#define SIMPLE_AP_FLAG_CAN_CONNECT   (1 << 0)
#define SIMPLE_AP_FLAG_QUICK_AUTH    (1 << 1)
#define SIMPLE_AP_FLAG_OPEN          (1 << 2)
#define SIMPLE_AP_FLAG_HIDDEN        (1 << 3)

#define SIMPLE_AP_FLAG_REASON_SHIFT              (4)
#define SIMPLE_AP_FLAG_GROUP_CIPHER_SHIFT        (8)
#define SIMPLE_AP_FLAG_PAIRWISE_CIPHER_SHIFT     (24)
#define SIMPLE_AP_FLAG_AKM_SHIFT                 (40)
#define SIMPLE_AP_FLAG_RSN_CAP_SHIFT             (56)

enum {
    SIMPLE_AP_FLAG_REASON_IBSS  = 1,  /* important that reason codes are non-zero */
    SIMPLE_AP_FLAG_REASON_8021X = 2,
    SIMPLE_AP_FLAG_REASON_MFPR  = 3,
    SIMPLE_AP_FLAG_REASON_RSSI  = 4
};

/* Copied and extended from atheros/driver/ath_beacon.h */
#define R_AUTH_8021X            (1 << 0)
#define R_AUTH_PSK              (1 << 1)
#define R_AUTH_FT_8021X         (1 << 2)
#define R_AUTH_FT_PSK           (1 << 3)
#define R_AUTH_SHA256_8021X     (1 << 4)
#define R_AUTH_SHA256_PSK       (1 << 5)
#define R_AUTH_TDLS             (1 << 6)
#define R_AUTH_SAE              (1 << 7)
#define R_AUTH_FT_SAE           (1 << 8)
#define R_AUTH_AP_PEER_KEY      (1 << 9)
#define R_AUTH_SHA256_8021X_B   (1 << 10)
#define R_AUTH_SHA384_8021X     (1 << 11)
#define R_AUTH_FT_SHA384_8021X  (1 << 12)

#define R_CIPHER_USE_GROUP              (1 << 0)
#define R_CIPHER_WEP_40                 (1 << 1)
#define R_CIPHER_TKIP                   (1 << 2)
#define R_CIPHER_CCMP_128               (1 << 3)
#define R_CIPHER_WEP_104                (1 << 4)
#define R_CIPHER_BIP_CMAC_128           (1 << 5)
#define R_CIPHER_GCMP_128               (1 << 6)
#define R_CIPHER_GCMP_256               (1 << 7)
#define R_CIPHER_CCMP_256               (1 << 8)
#define R_CIPHER_BIP_GMAC_128           (1 << 9)
#define R_CIPHER_BIP_GMAC_256           (1 << 10)
#define R_CIPHER_BIP_CMAC_256           (1 << 11)

#define R_RSN_CAP_MFPR                  (1 << 0)
#define R_RSN_CAP_MFPC                  (1 << 1)

#define CAP_INFO_PRIVACY_BIT (1 << 4)

/* ------------------------------------------------------------------------------------------------------- */
/* Link Statistics */

#define ATH_MAX_TXQS_TO_REPORT 10

#if !defined(IEEE80211_ADDR_LEN)
#define IEEE80211_ADDR_LEN 6
#endif

#if !defined(IEEE80211_SSID_LEN)
#define IEEE80211_SSID_LEN 32
#endif

/*
 * Max # of links to allow, should be same value as NL_MAX_NEIGHBORS
 */
#define ATH_MAX_WIRELESS_LINKS 32

struct simple80211_link_stats {
    unsigned short  model;
    unsigned short  submodel;
    unsigned char   macaddr[IEEE80211_ADDR_LEN];
    unsigned char   per;
    unsigned char   rssiFrom;
    unsigned char   rssiTo;
    unsigned int    pktsMin;
};

struct simple80211_ap_disconnect_stats {
    unsigned int  deAuthCount;
    unsigned int  deAuthReasonCodeMap;
    unsigned long deAuthMaxReconnectTime;
};

#if defined(SONOS_ARCH_ATTR_SUPPORTS_EXTERNAL_STATION) || defined(SONOS_ARCH_ATTR_USES_DRIVER_NL80211)

#if !defined(WMM_NUM_OF_AC)
#define WMM_NUM_OF_AC   (4)
#endif

struct simple80211_edca_param {
    unsigned char  is_valid;
    unsigned char  aifsn[WMM_NUM_OF_AC];
    unsigned char  cwmin[WMM_NUM_OF_AC];
    unsigned char  cwmax[WMM_NUM_OF_AC];
    unsigned short txop[WMM_NUM_OF_AC];
    unsigned char  acm[WMM_NUM_OF_AC];
};
#endif

#define SIMPLE80211_AP_STATS_KEY_MGMT_STRING_LEN 5
#define SIMPLE80211_AP_STATS_CIPHER_STRING_LEN 5

struct simple80211_ap_stats {
    unsigned char bssid[IEEE80211_ADDR_LEN];
    unsigned char perToAP;
    unsigned char rssiAP;
    unsigned int  pktsMin;
    struct simple80211_ap_disconnect_stats dStats;
    unsigned int  totalDataRx;
    unsigned int  totalSecDrops;
    unsigned int  wpaReplayDrops;
    unsigned char apStatToReport;
    unsigned char ssid[IEEE80211_SSID_LEN];
    unsigned char ssidLen;
    unsigned int  rxPrr;
    unsigned char keyMgmt[SIMPLE80211_AP_STATS_KEY_MGMT_STRING_LEN];
    unsigned char peerCipher[SIMPLE80211_AP_STATS_CIPHER_STRING_LEN];
    unsigned char GroupCipher[SIMPLE80211_AP_STATS_CIPHER_STRING_LEN];
#if defined(SONOS_ARCH_ATTR_SUPPORTS_EXTERNAL_STATION) || defined(SONOS_ARCH_ATTR_USES_DRIVER_NL80211)
    unsigned int  txLatency;
    struct simple80211_edca_param edca;
#endif
};

struct simple80211_txq_stats {
    unsigned int qID;
    unsigned int noBuff;
    unsigned int completed;
    unsigned int allocBfPeak;
    unsigned int latency;
};

struct simple80211_ani_stats {
    unsigned int ani_l;
};

struct simple80211_mibcc_stats {
    unsigned int rxFramePct;    // % cycles receiving a frame
    unsigned int rxClearPct;    // % cycles Rx is clear (medium busy)
    unsigned int txFramePct;    // % cycles transmitting a frame
};

// We should rename following structure to something like player_stats_info
struct simple80211_link_stats_info {
    unsigned int freq;
    unsigned int phyErrCount;
    // Noise floor reporting
    int nf_0;
    int nf_1;
    int nf_2;
    /*
       Unless we want to report following counters in SonosNet mode
       we should move them to ap_stats above
    */
    unsigned int gtkFailMode;   // non-zero if actively failing (for any
                                // reason, not just gtk anymore).  The name is
                                // unfortunate, but I'd rather not force the
                                // Splunk guys to change and it makes sense to
                                // have the name here match the Splunk name.
    unsigned int gtkFailCount;  // read/clear counter of gtk failures
    unsigned int ngcFailCount;  // read/clear counter of ngc failures
    unsigned int latencyBelow;
    unsigned int latencyAbove;
    unsigned int latencyAvg;
    unsigned int numLinkStats;
    unsigned int txqStatsUpdated;
    unsigned int rxqCompletedPeak;
    struct simple80211_link_stats linkStats[ATH_MAX_WIRELESS_LINKS];
    struct simple80211_ap_stats apStats;
    struct simple80211_txq_stats txqStats[ATH_MAX_TXQS_TO_REPORT];
    struct simple80211_ani_stats aniStats;
    struct simple80211_mibcc_stats mibcc;
};

#endif
